import re
import monkeypatch

from django.apps import AppConfig
from django.http import HttpRequest
from django.conf.urls import url, include

from .utils import from_dotted_path, HttpRequest__get_host
from .app_settings import app_settings

class DynamicSubdomainsConfig(AppConfig):
    name = 'dynamic_subdomains'

    def ready(self):
        for x in app_settings.SUBDOMAINS:
            # We add a literal period to the end of every pattern to avoid rather
            # unwieldy escaping in every definition.
            x['_regex'] = re.compile(r'%s(\.|$)' % x['regex'])
            x['_callback'] = from_dotted_path(x['callback'])

        if app_settings.EMULATE:
            monkeypatch.patch(HttpRequest__get_host, HttpRequest, 'get_host')

            # Inject our URLs
            for x in app_settings.SUBDOMAINS:
                urlconf_module = from_dotted_path(x['urlconf'])
                urlconf_module.urlpatterns = list(urlconf_module.urlpatterns) + [
                    url(r'^_/subdomains/', include('dynamic_subdomains.urls',
                        namespace='dynamic-subdomains')),
                ]



import contextlib

from django.core.urlresolvers import set_urlconf
from django.core.urlresolvers import NoReverseMatch

from .app_settings import app_settings

@contextlib.contextmanager
def set_urlconf_from_host(host):
    # Find best match, falling back to DEFAULT_SUBDOMAIN
    for subdomain in app_settings.SUBDOMAINS:
        match = subdomain['_regex'].match(host)
        if match:
            kwargs = match.groupdict()
            break
    else:
        kwargs = {}
        subdomain = get_subdomain(app_settings.DEFAULT_SUBDOMAIN)

    set_urlconf(subdomain['urlconf'])

    try:
        yield subdomain, kwargs
    finally:
        set_urlconf(None)

def get_subdomain(name):
    try:
        return {x['name']: x for x in app_settings.SUBDOMAINS}[name]
    except KeyError:
        raise NoReverseMatch("No subdomain called %s exists" % name)

def from_dotted_path(fullpath):
    """
    Returns the specified attribute of a module, specified by a string.

    ``from_dotted_path('a.b.c.d')`` is roughly equivalent to::

        from a.b.c import d

    except that ``d`` is returned and not entered into the current namespace.
    """

    module, attr = fullpath.rsplit('.', 1)

    return getattr(
        __import__(module, {}, {}, (attr,)),
        attr,
    )

def HttpRequest__get_host(self, *args, **kwargs):
    try:
        return self.COOKIES[app_settings.COOKIE_NAME]
    except KeyError:
        return HttpRequest__get_host._get_host(self, *args, **kwargs)

def noop(*args, **kwargs):
    return



from django.conf import settings

def setting(name, default):
    @property
    def fn(self):
        return getattr(settings, name, default)
    return fn

class AppSettings(object):
    # Required
    SUBDOMAINS = setting('SUBDOMAINS', ())
    DEFAULT_SUBDOMAIN = setting('DEFAULT_SUBDOMAIN', None)

    PROTOCOL = setting('SUBDOMAINS_PROTOCOL', 'http')
    EMULATE_BASE_URL = setting('EMULATE_BASE_URL', 'http://127.0.0.1:8000')

    EMULATE = setting('EMULATE_SUBDOMAINS', settings.DEBUG)
    COOKIE_NAME = setting('SUBDOMAINS_COOKIE_NAME', '_dynamic_subdomains_host')

app_settings = AppSettings()



class subdomain(dict):
    def __init__(self, regex, urlconf, name, callback='dynamic_subdomains.utils.noop'):
        self.update({
            'regex': regex,
            'urlconf': urlconf,
            'name': name,
            'callback': callback,
        })



from django.conf import settings
from django.template import Engine, Context
from django.utils.encoding import smart_unicode
from django.core.exceptions import MiddlewareNotUsed

from .utils import set_urlconf_from_host
from .app_settings import app_settings

class SubdomainMiddleware(object):
    """
    Adjust incoming request's urlconf based on `settings.SUBDOMAINS`.

    Overview
    ========

    This middleware routes requests to specific subdomains to different URL
    schemes ("urlconf").

    For example, if you own ``example.com`` but want to serve specific content
    at ``api.example.com` and ``beta.example.com``, add the following to your
    ``settings.py``:

        from dynamic_subdomains.subdomains import subdomain

        SUBDOMAINS = (
            subdomain(r'api', 'path.to.api.urls',
                name='api'),
            subdomain(r'beta', 'path.to.beta.urls',
                name='beta'),
        )

    This causes requests to ``{api,beta}.example.com`` to be routed to their
    corresponding urlconf. You can use your ``urls.py`` as a template for these
    urlconfs.

    Patterns are evaluated in order. If no subdomain pattern matches, the
    subdomain specified by ``settings.DEFAULT_SUBDOMAIN`` is used.

    Pattern format
    ==============

    The patterns on the left-hand side are regular expressions. For example,
    the following ``settings.SUBDOMAINS`` will route ``foo.example.com`` and
    ``bar.example.com`` to the same urlconf.

        SUBDOMAINS = (
            subdomain(r'(foo|bar)', 'path.to.urls',
                name='foo-or-bar'),
        )

    .. note:

      * Patterns are matched against the extreme left of the requested host

      * It is implied that all patterns end either with a literal full stop
        (ie. ".") or an end of line metacharacter.

      * As with all regular expressions, various metacharacters need quoting.

    Using regular expressions
    =========================

    Patterns as regular expressions allows setups to feature dynamic (or
    "wildcard") subdomain schemes:

        SUBDOMAINS = (
            subdomain(r'www', ROOT_URLCONF,
                name='static'),
            subdomain(r'\w+', 'path.to.custom_urls',
                name='wildcard'),
        )

    Here, requests to ``www.example.com`` will be routed as normal but a
    request to ``lamby.example.com`` is routed to ``path.to.custom_urls``.

    As subdomains are matched in order, we placed ``www`` first as it otherwise
    would have matched against ``\w+`` and thus routed to the wrong
    destination.

    Alternatively, we could have used negative lookahead:

        SUBDOMAINS = (
            subdomain(r'(?!www)\w+', 'path.to.custom_urls',
                name='wildcard'),
        )

    Callback methods to simplify dynamic subdomains
    ===============================================

    The previous section outlined using regular expressions to implement
    dynamic subdomains.

    However, inside every view referenced by the target urlconf we would have
    to parse the subdomain from ``request.get_host()`` and lookup its
    corresponding object instance, violating DRY. If these dynamic subdomains
    had a lot of views this would become particularly unwieldy.

    To remedy this, you can optionally specify a callback method to be called
    if your subdomain matches:

        SUBDOMAINS = (
            subdomain(r'www', ROOT_URLCONF,
                name='static'),
            subdomain(r'(?P<username>\w+)', 'path.to.custom_urls',
                callback='path.to.custom_fn', name='with-callback'),
        )

        [..]

        from django.shortcuts import get_object_or_404
        from django.contrib.auth.models import User

        def custom_fn(request, username):
            request.viewing_user = get_object_or_404(User, username=username)

    This example avoids the duplicated work in every view by attaching a
    ``viewing_user`` instance to the request object. Views referenced by the
    "dynamic" urlconf can now assume that this object exists.

    The custom method is called with the ``request`` object and any named
    captured arguments, similar to regular Django url processing.

    Callbacks may return either ``None`` or an ``HttpResponse`` object. If it
    returns ``None``, the request continues to be processed and the appropriate
    view is eventually called. If a callback returns an ``HttpResponse``
    object, that ``HttpResponse`` is returned to the client without any further
    processing.

    .. note:

        Callbacks are executed with the urlconf set to the second argument in
        the ``SUBDOMAINS`` list. For example, in the example above, the
        callback will be executed with the urlconf as ``path.to.custom_urls``
        and not the default urlconf.

        This can cause problems when reversing URLs within your callback as
        they may not be "visible" to ``django.core.urlresolvers.reverse`` as
        they are specified in (eg.) the default urlconf.

        To remedy this, specify the ``urlconf`` parameter when calling
        ``reverse``.

    Notes
    =====

      * When using dynamic subdomains based on user input, ensure users cannot
        specify names that conflict with static subdomains such as "www" or
        their subdomain will not be accessible.

      * Don't forget to add ``handler404`` and ``handler500`` entries for your
        custom urlconfs.
    """

    TAG = '</body>'
    TEMPLATE = """
        <script type="text/javascript">
            function dynamicSubdomains() {
                var val = prompt("Update host", '{{ request.get_host|escapejs }}');
                document.cookie = '{{ app_settings.COOKIE_NAME }}=' + val + '; expires=Tue, 19 Jan 2038 03:14:07 UTC;';
                window.location.reload();
            }
        </script>
        <style type="text/css">
            #dynamic-subdomains-toolbar {
                right: 0;
                bottom: 0;
                height: 36px;
                z-index: 99999;
                position: fixed;

                padding: 10px 16px 0 16px;
                background-color: #092e20;
                border-top-left-radius: 7px;

                color: white;
                cursor: pointer;
                font-size: 12px;
                font-weight: bold;
            }
        </style>
        <div id="dynamic-subdomains-toolbar" onclick="dynamicSubdomains();">
            {{ request.build_absolute_uri }}
        </div>
    """

    def __init__(self):
        if not settings.SUBDOMAINS:
            raise MiddlewareNotUsed()

    def process_request(self, request):
        host = request.get_host()

        with set_urlconf_from_host(host) as (subdomain, kwargs):
            request.urlconf = subdomain['urlconf']

            return subdomain['_callback'](request, **kwargs)

    def process_response(self, request, response):
        if not app_settings.EMULATE:
            return response

        try:
            val = smart_unicode(response.content)
        except Exception:
            return response

        toolbar = Engine(debug=True).from_string(self.TEMPLATE).render(Context({
            'request': request,
            'app_settings': app_settings,
        }))

        idx = val.lower().rfind(self.TAG.lower())

        if idx >= 0:
            response.content = val[:idx] + toolbar + val[idx + len(self.TAG):]

        return response



from django.conf import settings
from django.http import HttpResponseRedirect, HttpResponseForbidden

from .app_settings import app_settings

def redirect_(request, host):
    if not settings.DEBUG:
        return HttpResponseForbidden()

    response = HttpResponseRedirect(request.META.get('QUERY_STRING', '') or '/')
    response.set_cookie(app_settings.COOKIE_NAME, host)
    response.status_code = 307 # Re-submit POST requests

    return response



#!/usr/bin/env python

from setuptools import setup, find_packages

setup(
    name='django-dynamic-subdomains',
    description="Dynamic and static subdomain support for Django.",
    version='1.0.0',
    url='https://chris-lamb.co.uk/projects/django-dynamic-subdomains/',

    author="Chris Lamb",
    author_email='chris@chris-lamb.co.uk',
    license='BSD',

    packages=find_packages(),

    install_requires=(
        'Django>=1.8',
        'monkeypatch==0.1rc3',
    ),
)



from django import template
from django.template import TemplateSyntaxError
from django.utils.encoding import smart_str
from django.template.defaulttags import kwarg_re

from ..reverse import reverse_crossdomain
from ..app_settings import app_settings

register = template.Library()

@register.tag
def domain_url(parser, token):
    bits = token.split_contents()
    if len(bits) < 2:
        raise TemplateSyntaxError("'%s' takes at least 1 argument" % bits[0])

    view = parser.compile_filter(bits[1])
    bits = bits[1:] # Strip off view

    try:
        pivot = bits.index('on')
    except ValueError:
        # No "on <subdomain>" was specified so use the default domain
        domain = app_settings.DEFAULT_SUBDOMAIN
        view_args, view_kwargs = parse_args_kwargs(parser, bits[1:])
        domain_args, domain_kwargs = (), {}
    else:
        try:
            domain = bits[pivot + 1]
        except IndexError:
            raise TemplateSyntaxError(
                "'%s' arguments must include a domain after 'on'" % bits[0]
            )

        view_args, view_kwargs = parse_args_kwargs(parser, bits[1:pivot])
        domain_args, domain_kwargs = parse_args_kwargs(parser, bits[pivot + 2:])

    return DomainURLNode(
        domain,
        view,
        domain_args,
        domain_kwargs,
        view_args,
        view_kwargs,
    )

class DomainURLNode(template.Node):
    def __init__(self, subdomain, view, subdomain_args, subdomain_kwargs, view_args, view_kwargs):
        self.subdomain = subdomain
        self.view = view

        self.subdomain_args = subdomain_args
        self.subdomain_kwargs = subdomain_kwargs

        self.view_args = view_args
        self.view_kwargs = view_kwargs

    def render(self, context):
        subdomain_args = [x.resolve(context) for x in self.subdomain_args]
        subdomain_kwargs = {
            smart_str(k, 'ascii'): v.resolve(context)
            for k, v in self.subdomain_kwargs.items()
        }

        view_args = [x.resolve(context) for x in self.view_args]
        view_kwargs = {
            smart_str(k, 'ascii'): v.resolve(context)
            for k, v in self.view_kwargs.items()
        }

        return reverse_crossdomain(
            self.subdomain,
            self.view.resolve(context),
            subdomain_args,
            subdomain_kwargs,
            view_args,
            view_kwargs,
        )

def parse_args_kwargs(parser, bits):
    args = []
    kwargs = {}

    for bit in bits:
        match = kwarg_re.match(bit)
        if not match:
            raise TemplateSyntaxError("Malformed arguments to domain_url tag")

        name, value = match.groups()
        if name:
            kwargs[name] = parser.compile_filter(value)
        else:
            args.append(parser.compile_filter(value))

    return args, kwargs



import re
import urllib

from django.utils.encoding import force_unicode
from django.core.urlresolvers import NoReverseMatch, reverse
from django.utils.regex_helper import normalize

from .utils import get_subdomain
from .app_settings import app_settings

def reverse_subdomain(subdomain_name, args=(), kwargs=None):
    if args and kwargs:
        raise ValueError("Don't mix *args and **kwargs in call to reverse()!")

    if kwargs is None:
        kwargs = {}

    subdomain = get_subdomain(subdomain_name)

    unicode_args = [force_unicode(x) for x in args]
    unicode_kwargs = dict([(k, force_unicode(v)) for (k, v) in kwargs.items()])

    for result, params in normalize(subdomain['regex']):
        if args:
            if len(args) != len(params):
                continue
            candidate = result % dict(zip(params, unicode_args))
        else:
            if set(kwargs.keys()) != set(params):
                continue
            candidate = result % unicode_kwargs

        if re.match(subdomain['regex'], candidate, re.UNICODE):
            return candidate

    raise NoReverseMatch(
        "Reverse subdomain for '%s' with arguments '%s' and keyword arguments "
        "'%s' not found." % (subdomain_name, args, kwargs)
    )

def reverse_crossdomain_part(subdomain_name, path, subdomain_args=(), subdomain_kwargs=None):
    if subdomain_kwargs is None:
        subdomain_kwargs = {}

    domain_part = reverse_subdomain(
        subdomain_name,
        args=subdomain_args,
        kwargs=subdomain_kwargs,
    )

    if app_settings.EMULATE:
        url = '%s%s' % (
            app_settings.EMULATE_BASE_URL,
            reverse('dynamic-subdomains:redirect', args=(domain_part,)),
        )

        if path != '/':
            url += '?%s' % urllib.quote(path)

        return url

    return u'//%s%s' % (domain_part, path)

def reverse_path(subdomain_name, view, args=(), kwargs=None):
    if kwargs is None:
        kwargs = {}

    subdomain = get_subdomain(subdomain_name)

    return reverse(
        view,
        args=args,
        kwargs=kwargs,
        urlconf=subdomain['urlconf'],
    )

def reverse_crossdomain(subdomain_name, view, subdomain_args=(), subdomain_kwargs=None, view_args=(), view_kwargs=None):
    path = reverse_path(subdomain_name, view, view_args, view_kwargs)

    return reverse_crossdomain_part(
        subdomain_name,
        path,
        subdomain_args,
        subdomain_kwargs,
    )
